/**
 * Cecilia, an implementation of the Fractal component model in C.
 * 
 * Copyright (C) 2004 France Telecom R&D
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Contact: fractal@objectweb.org
 *
 * Authors: J.-P. Fassino, Matthieu Leclercq.
 */

#define IN_CECILIA_MECHANISM
#include "fractal/api/defaultComposite.h"
#include <cecilia.h>

// TODO add printf for DEBUG_COMPOSITE in composite start. 

#include <string.h>

extern int printf(const char * format, ...);

// checks a binding descriptor
static int checkBinding(struct __cecilia_defaultCompositeData_t *self,
    struct BindingDesc_t* bindingDesc);

// ensures that every server components of the given components are started.
static int startServers(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx, int allowedUAS);

// starts the given sub component.
static int startSubComp(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx);

// stops the given sub component.
static int stopSubComp(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx);

// Instantiates the given binding descriptor
static int bind(struct __cecilia_defaultCompositeData_t *self,
    struct BindingDesc_t* bindingDesc);

// get a interface of a sub component.
static int getItf(struct __cecilia_defaultCompositeData_t *self, int idx,
    char* name, void **itf)
{
  return CALL(self->subComponents[idx].instance, getFcInterface, name, itf);
}

// -----------------------------------------------------------------------------
// Implementation of the component interface.
// -----------------------------------------------------------------------------

/**
 * Implementation of the getFcInterface method.
 */
// int getFcInterface(in string interfaceName, out any interfaceReference);
static int getFcInterface(void* _this, char* interfaceName,
    void ** interfaceReference)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;

  int i;
  if (strcmp(interfaceName, "component") == 0) {
    *interfaceReference = &(self->myreference);
    return fractal_api_ErrorConst_OK;
  } else if (strcmp(interfaceName, "binding-controller") == 0) {
    *interfaceReference = &(self->bc);
    return fractal_api_ErrorConst_OK;
  } else if (strcmp(interfaceName, "lifecycle-controller") == 0) {
    *interfaceReference = &(self->lcc);
    return fractal_api_ErrorConst_OK;
  } else if (strcmp(interfaceName, "content-controller") == 0) {
    *interfaceReference = &(self->cc);
    return fractal_api_ErrorConst_OK;
  }

  // Getting exported interfaces owned by sub-components
  for (i = 0; i < self->nbBindings; i++) {
    if ((self->bindings[i].client == COMPOSITE_IDX)&& (strcmp(interfaceName,
        self->bindings[i].clientName) == 0)) {
      return getItf(self, self->bindings[i].server,
          self->bindings[i].serverName, interfaceReference);
    }
  }

  return fractal_api_ErrorConst_NO_SUCH_INTERFACE;
}

/**
 * Implementation of the getFcInterfaces method.
 */
// int getFcInterfaces(in any[] interfaceReferences);
static int getFcInterfaces(void* _this, void* interfaceReferences[])
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i, err;
  int nbItf = 4; // there is at least 4 control interfaces.

  if (interfaceReferences != NULL) {
    interfaceReferences[0] = &self->myreference;
    interfaceReferences[1] = &self->bc;
    interfaceReferences[2] = &self->lcc;
    interfaceReferences[3] = &self->cc;

    for (i = 0; i < self->nbBindings; i++) {
      if (self->bindings[i].client == COMPOSITE_IDX) {
        err = getItf(self, self->bindings[i].server,
            self->bindings[i].serverName, &(interfaceReferences[nbItf]));
        if (err != fractal_api_ErrorConst_OK)
          return err;
        nbItf ++;
      }
    }
  } else { // interfaceReferences == NULL
    for (i = 0; i < self->nbBindings; i++) {
      if (self->bindings[i].client == COMPOSITE_IDX) {
        nbItf ++;
      }
    }
  }

  return nbItf;
}

/** Declaration of the V-Table of this interface implementation. */
struct Mfractal_api_Component __cecilia_defaultComposite_ComponentMeths = {
    getFcInterface, getFcInterfaces };

// -----------------------------------------------------------------------------
// Implementation of the binding-controller interface.
// -----------------------------------------------------------------------------

/**
 * Implementation of the listFc method.
 */
// int listFc(in string[] clientItfNames);
static int listFc(void* _this, char* clientItfNames[])
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i;
  int nbItf = 0;

  if (clientItfNames != NULL) {
    for (i = 0; i < self->nbBindings; i++) {
      if (self->bindings[i].server == COMPOSITE_IDX) {
        clientItfNames[nbItf] = self->bindings[i].serverName;
        nbItf++;
      }
    }
  } else { // clientItfNames == NULL
    for (i = 0; i < self->nbBindings; i++) {
      if (self->bindings[i].server == COMPOSITE_IDX) {
        nbItf++;
      }
    }
  }

  return nbItf;
}

/**
 * Implementation of the lookupFc method.
 */
// int lookupFc(in string clientItfName, out any interfaceReference);
static int lookupFc(void* _this, char* clientItfName, void** interfaceReference)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i, err;

  if (interfaceReference == NULL || clientItfName == NULL) {
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

  for (i = 0; i < self->nbBindings; i++) {
    if (self->bindings[i].server == COMPOSITE_IDX && strcmp(clientItfName,
        self->bindings[i].serverName) == 0) {
      Rfractal_api_BindingController* subBC;
      err = getItf(self, self->bindings[i].client, "binding-controller",
          interfaceReference);
      if (err != fractal_api_ErrorConst_OK)
        return err;
      return CALL(subBC, lookupFc, self->bindings[i].clientName,
          interfaceReference);
    }
  }

  return fractal_api_ErrorConst_NO_SUCH_INTERFACE;
}

/**
 * Implementation of the bindFc method.
 */
// int bindFc(in string clientItfName, in any serverItf);
static int bindFc(void* _this, char* clientItfName, void* serverItf)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i, err;
  int result = fractal_api_ErrorConst_NO_SUCH_INTERFACE;

  if (clientItfName == NULL) {
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

  for (i = 0; i < self->nbBindings; i++) {
    if (self->bindings[i].server == COMPOSITE_IDX && strcmp(clientItfName,
        self->bindings[i].serverName) == 0) {
      Rfractal_api_BindingController * subBC;
      err= getItf(self, self->bindings[i].client, "binding-controller",
          (void **) &subBC);
      if (err != fractal_api_ErrorConst_OK)
        return err;
      // TODO check err
      err = CALL(subBC, bindFc, self->bindings[i].clientName, serverItf);
      if (err != fractal_api_ErrorConst_OK)
        return err;
      
      result = fractal_api_ErrorConst_OK;
    }
  }

  return result;
}

/**
 * Implementation of the unbindFc method.
 */
// int unbindFc(in string clientItfName);
static int unbindFc(void* _this, char* clientItfName)
{
  return bindFc(_this, clientItfName, NULL);
}

struct Mfractal_api_BindingController
    __cecilia_defaultComposite_BindingControllerMeths = { listFc, lookupFc,
        bindFc, unbindFc };

// -----------------------------------------------------------------------------
// Implementation of the lifecycle-controller interface.
// -----------------------------------------------------------------------------

/**
 * Implementation of the getFcState method.
 */
// int getFcState();
static int getFcState(void* _this)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  if (self->started == 0)
    return fractal_api_LifeCycleController_STOPPED;
  else
    return fractal_api_LifeCycleController_STARTED;
}

// This bit is set on a component if has at least one "usedAtStart" client 
// interface. 
#define SUB_COMPONENT_HAS_UAS ((unsigned char) 0x10)

/**
 * Implementation of the startFc method.
 */
// int startFc();
static int startFc(void* _this)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i, state, err;
  int hasUsedAtStartBinding = 0;
  int hasBindAfterStartBinding = 0;

  // if already started, simply return
  if (self->started != 0) {
    return fractal_api_ErrorConst_OK;
  }

  // first create each binding
  for (i = 0; i < self->nbBindings; i++) {

    // check binding
    err = checkBinding(self, &(self->bindings[i]));
    if (err != fractal_api_ErrorConst_OK) {
      return err;
    }

    // if it is an import or an export binding, pass it.
    if (self->bindings[i].client == COMPOSITE_IDX|| self->bindings[i].server
        == COMPOSITE_IDX)
      continue;

    state = self->bindings[i].state;
    // if the binding is already done, pass it.
    if (GET_BINDING_STATUS(state) == BINDING_BOUND) {
      continue;
    }

    // if the binding must be done after the component start, 
    // note that there are "bind after start" binding and pass it.
    if (GET_BINDING_START_PROPERTY(state) == BINDING_BIND_AFTER_START) {
      hasBindAfterStartBinding = 1;
      continue;
    }

    // if the binding is used at start, note it and continue.
    if (GET_BINDING_START_PROPERTY(state) == BINDING_USED_AT_START) {
      hasUsedAtStartBinding = 1;
      self->subComponents[self->bindings[i].client].state
          |= SUB_COMPONENT_HAS_UAS;
    }

    // perform the actual binding.
    err = bind(self, &(self->bindings[i]));
    if (err != fractal_api_ErrorConst_OK)
      return err;

  } // end for each binding.

  if (hasUsedAtStartBinding) {
    // there are some "usedAtStart" bindings. must start the sub components in 
    // the correct order.

    // for each sub-component
    for (i = 0; i < self->nbSubComponents; i++) {

      if ((self->subComponents[i].state & SUB_COMPONENT_HAS_UAS) == 0) {
        // the component is not known as having "usedAtStart" client interface
        // simply start it.
        err = startSubComp(self, i);
        if (err != fractal_api_ErrorConst_OK)
          return err;

      } else {
        // the component has a "usedAtstart" client interface.
        startServers(self, i, i);
      }
    } // end for each sub component

    // clear the SUB_COMPONENT_HAS_UAS bits
    for (i = 0; i < self->nbSubComponents; i++) {
      self->subComponents[i].state &= ~SUB_COMPONENT_HAS_UAS;
    }

  } else {
    // there is no "usedAtStart" bindings.
    for (i = 0; i < self->nbSubComponents; i++) {
      err = startSubComp(self, i);
      if (err != fractal_api_ErrorConst_OK)
        return err;
    }
  } // end if (hasUsedAtStartBinding)

  if (hasBindAfterStartBinding) {
    // there are some 'bindAfterStart' binding to bind.

    // for each binding
    for (i = 0; i < self->nbBindings; i++) {

      // if the binding is 'bindAfterStart'
      if (GET_BINDING_START_PROPERTY(self->bindings[i].state)
          == BINDING_BIND_AFTER_START) {

        // bind it.
        err = bind(self, &(self->bindings[i]));
        if (err != fractal_api_ErrorConst_OK)
          return err;
      }
    } // end for each binding
  } // end if (hasBindAfterStartBinding)

  self->started = 1;
  return fractal_api_ErrorConst_OK;

} // end startFc method


/**
 * Implementation of the stopFc method.
 */
// int stopFc();
static int stopFc(void* _this)
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i, err;

  for (i = 0; i < self->nbSubComponents; i++) {
    err = stopSubComp(self, i);
    if (err != fractal_api_ErrorConst_OK)
      return err;
  }

  return fractal_api_ErrorConst_OK;
}

struct Mfractal_api_LifeCycleController
    __cecilia_defaultComposite_LifeCycleControllerMeths = { getFcState,
        startFc, stopFc };

// -----------------------------------------------------------------------------
// Implementation of the content-controller interface.
// -----------------------------------------------------------------------------

/**
 * Implementation of the getFcSubComponents method.
 */
// int getFcSubComponents(in Component[] subComponents)
static int getFcSubComponents(void *_this,
    Rfractal_api_Component* subComponents[])
{
  struct __cecilia_defaultCompositeData_t
      *self = (struct __cecilia_defaultCompositeData_t *) _this;
  int i;

  if (subComponents != NULL) {
    for (i = 0; i < self->nbSubComponents; i++) {
      subComponents[i] = self->subComponents[i].instance;
    }
  }

  return self->nbSubComponents;
}

/**
 * Implementation of the addFcSubComponent method (not supported).
 */
// int addFcSubComponent(in Component subComponent)
static int addFcSubComponent(void *_this, Rfractal_api_Component* subComponent)
{
  return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;
}

/**
 * Implementation of the removeFcSubComponent method (not supported).
 */
// int removeFcSubComponent(in Component subComponent)
static int removeFcSubComponent(void *_this,
    Rfractal_api_Component* subComponent)
{
  return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;
}

/**
 * Implementation of the addFcSubBinding method (not supported).
 */
// int addFcSubBinding(in Component clientComponent, in string clientItfName,
//     in Component serverComponent, in string serverItfName)
static int addFcSubBinding(void *_this,
    Rfractal_api_Component* clientComponent, char* clientItfName,
    Rfractal_api_Component* serverComponent, char* serverItfName)
{
  return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;
}

/**
 * Implementation of the removeFcSubBinding method (not supported).
 */
// int removeFcSubBinding(in Component clientComponent, in string clientItfName)
static int removeFcSubBinding(void *_this,
    Rfractal_api_Component* clientComponent, char* clientItfName)
{
  return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;
}

struct Mfractal_api_ContentController
    __cecilia_defaultComposite_ContentControllerMeths = { getFcSubComponents,
        addFcSubComponent, removeFcSubComponent, addFcSubBinding,
        removeFcSubBinding };

// -----------------------------------------------------------------------------
// startServers utility method implementation
// -----------------------------------------------------------------------------

// this bit is set on components by 'startClients' to detect cycle in component
// graph.
#define SUB_COMPONENT_IN_STARTS_SERVER ((unsigned char) 0x20)

/**
 * This method performs a deep-first walk in the composition graph to start 
 * every servers of the component with index 'subCompIdx'.
 * This method use the 'SUB_COMPONENT_IN_STARTS_SERVER' bit to detect cycle in
 * the graph and avoid infinit recursion.
 * The 'allowedUAS' parameter is the index of the component that is allowed to
 * be in a cycle with a 'usedAtStart' client interface. This parameter is used 
 * to ensures that in each cycle of the composition graph there is at most one 
 * 'usedAtStart' binding. 
 */
static int startServers(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx, int allowedUAS)
{
  int i, err, serverIdx, serverState;

  // first set IN_STARTS_CLIENT state
  self->subComponents[subCompIdx].state |= SUB_COMPONENT_IN_STARTS_SERVER;

  // for each internal (not "bindAfterStart") binding for which subCompIdx is 
  // the client.
  for (i = 0; i < self->nbBindings; i++) {
    if ((self->bindings[i].client != subCompIdx) || (self->bindings[i].server
        == COMPOSITE_IDX)
        || (GET_BINDING_START_PROPERTY(self->bindings[i].state)
            == BINDING_BIND_AFTER_START)) {
      continue;
    }

    serverIdx = self->bindings[i].server;
    serverState = self->subComponents[serverIdx].state;
    // if the server component has a "usedAtStart" client interface.
    if ((serverState & SUB_COMPONENT_HAS_UAS) != 0) {

      // if the server component is not IN_STARTS_SERVER
      if ((serverState & SUB_COMPONENT_IN_STARTS_SERVER) == 0) {

        // if the server component is not already started
        if (GET_SUB_COMPONENT_STATUS(serverState) != SUB_COMPONENT_STARTED) {
          // starts every servers of the server component.
          // the allowed UAS component is the server component.
          err = startServers(self, serverIdx, serverIdx);
          if (err != fractal_api_ErrorConst_OK)
            return err;
        }
        // if the server is already started, then every server components of
        // this server are also started. so it is not required to call 
        // startServers for this component.

      } else {
        // the server component is already IN_STARTS_SERVER. there is a cycle.
        if (serverIdx != allowedUAS) {
          // ERROR! cycle in start order. 
#ifdef DEBUGCOMPOSITE
          printf("Invalid composition graph: two 'usedAtStart' client interfaces are involved in a binding cycle comp1=%d, comp2=%d\n",
              allowedUAS, serverIdx);
#endif // DEBUGCOMPOSITE
          return fractal_api_ErrorConst_ILLEGAL_CONTENT;
        } else {
          // else the server component is the "allowed UAS" component, so pass.
          continue;
        }
      }
    } else {
      // if the server component is not already IN_STARTS_SERVER
      if ((serverState & SUB_COMPONENT_IN_STARTS_SERVER) == 0) {
        // starts every servers of the server component.
        err = startServers(self, serverIdx, allowedUAS);
        if (err != fractal_api_ErrorConst_OK)
          return err;
      }
    }
  }

  // every client components are correctly started.
  // starts the component
  err = startSubComp(self, subCompIdx);
  if (err != fractal_api_ErrorConst_OK)
    return err;

  // clear the IN_STARTS_SERVER state
  self->subComponents[subCompIdx].state &= ~SUB_COMPONENT_IN_STARTS_SERVER;

  return fractal_api_ErrorConst_OK;
}

// -----------------------------------------------------------------------------
// [start/stop]Subcomp utility method implementation
// -----------------------------------------------------------------------------

static int startSubComp(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx)
{
  int err;
  Rfractal_api_LifeCycleController* lcc;

  if (GET_SUB_COMPONENT_STATUS(self->subComponents[subCompIdx].state)
      != SUB_COMPONENT_STARTED) {
    err = getItf(self, subCompIdx, "lifecycle-controller", (void **) &lcc);
    if (err != fractal_api_ErrorConst_NO_SUCH_INTERFACE) {
      if (err != fractal_api_ErrorConst_OK)
        return err;

      // call startFc on the sub component.
      err = CALL(lcc, startFc);
      if (err != fractal_api_ErrorConst_OK)
        return err;
    }
    self->subComponents[subCompIdx].state = SUB_COMPONENT_STARTED
        |(self->subComponents[subCompIdx].state & (~SUB_COMPONENT_STATUS_MASK));
  }

  return fractal_api_ErrorConst_OK;
}

static int stopSubComp(struct __cecilia_defaultCompositeData_t *self,
    int subCompIdx)
{
  int err;
  Rfractal_api_LifeCycleController* lcc;

  if (GET_SUB_COMPONENT_STATUS(self->subComponents[subCompIdx].state)
      != SUB_COMPONENT_STOPPED) {
    err = getItf(self, subCompIdx, "lifecycle-controller", (void **) &lcc);
    if (err != fractal_api_ErrorConst_NO_SUCH_INTERFACE) {
      if (err != fractal_api_ErrorConst_OK)
        return err;

      // call startFc on the sub component.
      err = CALL(lcc, stopFc);
      if (err != fractal_api_ErrorConst_OK)
        return err;
    }
    self->subComponents[subCompIdx].state = SUB_COMPONENT_STOPPED
        |(self->subComponents[subCompIdx].state & (~SUB_COMPONENT_STATUS_MASK));
  }

  return fractal_api_ErrorConst_OK;
}

// -----------------------------------------------------------------------------
// bind utility method implementation
// -----------------------------------------------------------------------------

static int bind(struct __cecilia_defaultCompositeData_t *self,
    struct BindingDesc_t* bindingDesc)
{
  Rfractal_api_BindingController* clientBc;
  void* serverItf;
  int err;

  // First step, get the server interface
  err = CALL(self->subComponents[bindingDesc->server].instance, getFcInterface,
      bindingDesc->serverName, &serverItf);
  if (err != fractal_api_ErrorConst_OK) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: can't find server interface serverIdx=%d, name=%s, err=%d\n",
        bindingDesc->server, bindingDesc->serverName, err);
#endif // DEBUGCOMPOSITE
    return err;
  }

  // Next get the BindingController interface of the client component
  err = CALL(self->subComponents[bindingDesc->client].instance, getFcInterface,
      "binding-controller", (void **) &clientBc);
  if (err != fractal_api_ErrorConst_OK) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: unknown can't find client binding controller clientIdx=%d, err=%d\n",
        bindingDesc->client, err);
#endif // DEBUGCOMPOSITE
    return err;
  }

  // Effective binding
  err = CALL2(clientBc, bindFc, bindingDesc->clientName, serverItf);
  if (err != fractal_api_ErrorConst_OK) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: error while binding client err=%d\n",
        err);
#endif // DEBUGCOMPOSITE
    return err;
  }

  return fractal_api_ErrorConst_OK;
}

// -----------------------------------------------------------------------------
// checkBinding utility method implementation
// -----------------------------------------------------------------------------

static int checkBinding(struct __cecilia_defaultCompositeData_t *self,
    struct BindingDesc_t* bindingDesc)
{
#ifdef CHECKCOMPOSITE

  // check the server index
  if (bindingDesc->server >= self->nbSubComponents) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: serverIdx=%d\n", bindingDesc->server);
#endif // DEBUGCOMPOSITE
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

  // check the client index
  if (bindingDesc->client >= self->nbSubComponents) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: clientIdx=%d\n", bindingDesc->client);
#endif // DEBUGCOMPOSITE
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

  // check the serverName
  if (bindingDesc->serverName == NULL) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: serverName is NULL\n");
#endif // DEBUGCOMPOSITE
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

  // check the clientName
  if (bindingDesc->clientName == NULL) {
#ifdef DEBUGCOMPOSITE
    printf("Invalid binding descriptor: clientName is NULL\n");
#endif // DEBUGCOMPOSITE
    return fractal_api_ErrorConst_INVALID_ARGUMENT;
  }

#endif // CHECKCOMPOSITE
  return fractal_api_ErrorConst_OK;
}
