/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 */

package org.objectweb.fractal.cecilia.adl.types;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.ContextLocal;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.components.PrimitiveCompiler;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.task.codegen.lib.AbstractFactoryProviderTask;
import org.objectweb.fractal.task.codegen.lib.AbstractInstanceProviderTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * This component determines if the component to compile is an instance of a
 * component definition that has already been compiled. If so, this component
 * does not call its {@link #definitionCompilerItf definition compiler} client
 * interface. <br>
 * <b>Note:</b> Here, we use "definition" in the sens of definition of the
 * <i>Cecilia type</i> of a component. <br>
 * In addition this component is responsible to add a
 * {@link org.objectweb.fractal.task.codegen.api.FactoryProviderTask "type" task} and a
 * {@link org.objectweb.fractal.task.codegen.api.InstanceProviderTask "instance" task}
 * which respectively defines the name of the <i>Cecilia type</i> of a component
 * and the name of the component instance.
 * 
 * @see ComponentDefinitionDesc
 */
public class PrimitiveDefinitionCompiler
    implements
      PrimitiveCompiler,
      BindingController {

  /**
   * A context local map associating {@link ComponentDefinitionDesc} and the
   * component node that define it.
   */
  protected final ContextLocal<Map<ComponentDefinitionDesc, ComponentContainer>> componentTypes          = new ContextLocal<Map<ComponentDefinitionDesc, ComponentContainer>>();

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /**
   * the name of the definition compiler that is executed when a new component
   * definition is detected.
   */
  public static final String                                                     DEFINITION_COMPILER_ITF = "definition-compiler";

  /**
   * The definition compiler that is executed when a new component definition is
   * detected.
   */
  public PrimitiveCompiler                                                       definitionCompilerItf;

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {

    try {
      // the task may already exist, in case of a shared component
      tasks.getTask("instance", container);
      // if so, simply return.
      return;
    } catch (final NoSuchElementException e) {
      // if not, continue.
    }

    // retrieve the component type map for the current context
    Map<ComponentDefinitionDesc, ComponentContainer> types = componentTypes
        .get(context);
    if (types == null) {
      types = new HashMap<ComponentDefinitionDesc, ComponentContainer>();
      componentTypes.set(context, types);
    }

    // compute component names
    final String fullName = getDefinitionName(path, container);
    final String instanceName = getInstanceName(path, container);

    // compute component definition descriptor
    final ComponentDefinitionDesc componentDefDesc = getComponentDefinitionDesc(
        path, container, context);

    // look for a component previously compiled with the same definition
    // descriptor.
    final ComponentContainer typeComp = types.get(componentDefDesc);

    if (typeComp == null) {
      // the component has an unknown type.
      // add it in the component type map
      types.put(componentDefDesc, container);

      // set the type decoration
      TypeDecorationUtil.setTypeDecoration(container, container);

      // add the type name provider task
      tasks.addTask("type", container, new TypeTask(fullName));

      // call the client definition compiler for this new component type
      definitionCompilerItf.compile(path, container, tasks, context);

    } else {
      // the component has a known type

      // set the type decoration
      TypeDecorationUtil.setTypeDecoration(container, typeComp);

      // find the type name provider task of the type
      final TypeTask typeTask = (TypeTask) tasks.getTask("type", typeComp);
      // add this task as the type name provider task of the component
      tasks.addTask("type", container, typeTask);

      /*
       * 'container' has the same set of interface as 'typeComp', but they may
       * not be in the same order. We must ensure that the 'container' component
       * and the 'typeComp' component have their interfaces in the same order.
       */
      if (container instanceof InterfaceContainer) {
        final Map<String, Interface> itfs = new HashMap<String, Interface>();
        // We first remove all the interfaces of 'container' and store
        // them in a map.
        final InterfaceContainer interfaceContainer = (InterfaceContainer) container;
        for (final Interface itf : interfaceContainer.getInterfaces()) {
          itfs.put(itf.getName(), itf);
          interfaceContainer.removeInterface(itf);
        }
        // Then re-add the stored interfaces respecting the order of 'typeComp'.
        for (final Interface itf : ((InterfaceContainer) typeComp)
            .getInterfaces()) {
          interfaceContainer.addInterface(itfs.get(itf.getName()));
        }
      }
    }

    // add the instance name provider task
    tasks.addTask("instance", container, new InstanceTask(instanceName));
  }

  protected ComponentDefinitionDesc getComponentDefinitionDesc(
      final List<ComponentContainer> path, final ComponentContainer container,
      final Map<Object, Object> context) {
    return new ComponentDefinitionDesc(container);
  }

  protected String getDefinitionName(final List<ComponentContainer> path,
      final ComponentContainer container) {
    String realname = getComponentName(container).replaceAll("-", "_");
    for (int i = path.size() - 1; i >= 0; i--)
      realname = getComponentName(path.get(i)).replaceAll("-", "_") + "_"
          + realname;
    return realname;
  }

  protected String getInstanceName(final List<ComponentContainer> path,
      final ComponentContainer container) {
    return getDefinitionName(path, container);
  }

  protected String getComponentName(final ComponentContainer container) {
    if (container instanceof Definition)
      return ((Definition) container).getName();
    else
      return ((Component) container).getName();

  }

  // --------------------------------------------------------------------------
  // Task classes
  // --------------------------------------------------------------------------

  protected class TypeTask extends AbstractFactoryProviderTask {

    protected TypeTask(final String typeName) {
      setFactory(typeName);
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      // nothing to do
    }
  }

  protected class InstanceTask extends AbstractInstanceProviderTask {

    protected InstanceTask(final String instanceName) {
      setInstance(instanceName);
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      // nothing to do
    }
  }

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public String[] listFc() {
    return new String[]{DEFINITION_COMPILER_ITF};
  }

  public void bindFc(final String itf, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {
    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(DEFINITION_COMPILER_ITF)) {
      this.definitionCompilerItf = (PrimitiveCompiler) value;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

  public Object lookupFc(final String itf) throws NoSuchInterfaceException {

    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(DEFINITION_COMPILER_ITF)) {
      return this.definitionCompilerItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

  public void unbindFc(final String itf) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {
    if (itf == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itf.equals(DEFINITION_COMPILER_ITF)) {
      this.definitionCompilerItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + itf
          + "'");
    }
  }

}
