/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.test;

import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getComplexType;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.isComplexType;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import junit.framework.TestCase;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.interfaces.IDLException;
import org.objectweb.fractal.adl.xml.XMLNodeFactoryImpl;
import org.objectweb.fractal.cecilia.adl.idl.BasicIDLLoader;
import org.objectweb.fractal.cecilia.adl.idl.ast.ComplexType;
import org.objectweb.fractal.cecilia.adl.idl.ast.Field;
import org.objectweb.fractal.cecilia.adl.idl.ast.FieldContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Import;
import org.objectweb.fractal.cecilia.adl.idl.ast.ImportContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinitionContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Method;
import org.objectweb.fractal.cecilia.adl.idl.ast.MethodContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Parameter;
import org.objectweb.fractal.cecilia.adl.idl.ast.ParameterContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinitionContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Type;
import org.objectweb.fractal.cecilia.adl.idl.exceptions.DuplicatedNameException;
import org.objectweb.fractal.cecilia.adl.idl.util.Util;

public class IDLLoaderTest extends TestCase {
  protected BasicIDLLoader loader;

  @Override
  protected void setUp() throws Exception {
    loader = new BasicIDLLoader();
    loader.nodeFactoryItf = new XMLNodeFactoryImpl();
  }

  public void testA1() throws Exception {
    loadInterface("A1", null);
  }

  public void testJ1() throws Exception {
    final InterfaceDefinition itf = loadInterface("J1", null);
    checkItfContent(itf, new String[]{"fieldI", "fieldJ"}, new String[]{
        "methodI", "methodJ"});
  }

  public void testJ2() throws Exception {
    try {
      loadInterface("J2", null);
      fail();
    } catch (final DuplicatedNameException e) {
      if (!(e.getSrc() instanceof Field)) fail();
      final Field f = (Field) e.getSrc();
      if (f == null) fail();
      if (!"fieldI".equals(f.getName())) fail();
    }
  }

  public void testJ3() throws Exception {
    try {
      loadInterface("J3", null);
      fail();
    } catch (final DuplicatedNameException e) {
      if (!(e.getSrc() instanceof Method)) fail();
      final Method m = (Method) e.getSrc();
      if (m == null) fail();
      if (!"methodI".equals(m.getName())) fail();
    }
  }

  public void testK() throws Exception {
    final InterfaceDefinition itf = loadInterface("K", null);
    checkItfContent(itf, new String[]{"fieldI", "fieldK"}, new String[]{
        "methodI", "methodK"});
  }

  public void testN() throws Exception {
    try {
      loadInterface("N", null);
      fail();
    } catch (final IDLException e) {
      if (e.getCause() == null) fail();
      if (!e.getCause().getMessage().contains("Cycle in inheritance graph"))
        fail();
    }
  }

  public void testQ() throws Exception {
    final InterfaceDefinition itf = loadInterface("Q", null);
    checkItfContent(itf, new String[]{"fieldP", "fieldQ"}, new String[]{
        "methodP", "methodQ"});
  }

  public void testErroneousI() throws Exception {
    loadErroneousIDL("ErroneousI1");
    loadErroneousIDL("ErroneousI2");
    loadErroneousIDL("ErroneousI3");
    loadErroneousIDL("ErroneousI4");
    loadErroneousIDL("ErroneousI5");
  }

  protected InterfaceDefinition loadInterface(final String name,
      final Map<Object, Object> ctx) throws Exception {
    return checkInterface((IDLDefinition) loader.loadInterface(
        "test.idl.interfaces." + name, (ctx == null)
            ? new HashMap<Object, Object>()
            : ctx), new HashSet<String>());
  }

  protected void loadErroneousIDL(final String name) throws Exception {
    try {
      loadInterface(name, null);
      fail("Loading of IDL '" + name + "' is suppose to fail");
    } catch (final IDLException e) {
      // OK
    }
  }

  // =====================================================================
  // utility methods to assert results

  protected static void checkItfContent(final InterfaceDefinition itf,
      final String[] fields, final String[] methods) {
    final Field[] fieldNodes = ((FieldContainer) itf).getFields();
    if (fields == null) {
      if (fieldNodes.length != 0) fail();
    } else {
      if (fieldNodes.length != fields.length)
        fail("The InterfaceDefinition does not contain the expected number of fields ("
            + fieldNodes.length + " instead of " + fields.length + ")");

      final Set<String> presentFields = new HashSet<String>(fieldNodes.length);
      for (final Field f : fieldNodes) {
        presentFields.add(f.getName());
      }

      for (final String f : fields) {
        if (!presentFields.contains(f)) {
          fail("The InterfaceDefinition does not contain a field named '" + f
              + "'");
        }
      }
    }

    final Method[] methodnodes = ((MethodContainer) itf).getMethods();
    if (methods == null) {
      if (methodnodes.length != 0) fail();
    } else {
      if (methodnodes.length != methods.length)
        fail("The InterfaceDefinition does not contains the expected number of methods ("
            + methodnodes.length + " instead of " + methods.length + ")");
      final Set<String> presentMethods = new HashSet<String>(methodnodes.length);

      for (final Method m : methodnodes) {
        presentMethods.add(m.getName());
      }

      for (final String methodToCheck : methods) {
        if (!presentMethods.contains(methodToCheck)) {
          fail("The InterfaceDefinition does not provide a method name '"
              + methodToCheck + "'");
        }
      }
    }
  }

  protected static InterfaceDefinition checkInterface(final IDLDefinition idl,
      final Set<String> checkedType) {
    if (!(idl instanceof InterfaceDefinition)) fail();
    final InterfaceDefinition itf = (InterfaceDefinition) idl;

    // check if it has already been checked
    if (!checkedType.add(idl.getName())) return itf;

    checkImportContainer((ImportContainer) itf);

    for (final Field field : ((FieldContainer) itf).getFields()) {
      final Type type = Util.getContainedType(field);
      if (Util.isComplexType(type)) {
        checkComplexType(Util.getComplexType(type), checkedType);
      }
      if (field.getName() == null) fail();
      if (field.getValue() == null) fail();
    }

    for (final Method m : ((MethodContainer) itf).getMethods()) {
      final Type returnType = Util.getContainedType((m));
      if (isComplexType(returnType))
        checkComplexType(Util.getComplexType(returnType), checkedType);
      if (m.getName() == null) fail();
      for (final Parameter p : ((ParameterContainer) m).getParameters()) {
        final Type parameterType = Util.getContainedType((p));
        if (isComplexType(parameterType))
          checkComplexType(getComplexType(parameterType), checkedType);
        if (p.getName() == null) fail();
        if (p.getQualifier() == null) fail();
      }
    }
    return itf;
  }

  protected static RecordDefinition checkRecord(final IDLDefinition idl,
      final Set<String> checkedType) {
    if (!(idl instanceof RecordDefinition)) fail();
    final RecordDefinition rec = (RecordDefinition) idl;

    // check if it has already been checked
    if (!checkedType.add(idl.getName())) return rec;

    checkImportContainer((ImportContainer) rec);

    for (final Field field : ((FieldContainer) rec).getFields()) {
      final Type type = Util.getContainedType(field);
      if (Util.isComplexType(type)) {
        checkComplexType(Util.getComplexType(type), checkedType);
      }
      if (field.getName() == null) fail();
      if (field.getValue() != null) fail();
    }
    return rec;
  }

  protected static void checkImportContainer(
      final ImportContainer importContainer) {
    for (final Import i : importContainer.getImports()) {
      final IDLDefinition impIDL = (IDLDefinition) ((Node) i)
          .astGetDecoration("idl-definition");
      if (impIDL == null) fail();
      if (!i.getName().equals(impIDL.getName())) fail();
    }
  }

  protected static void checkComplexType(final ComplexType ct,
      final Set<String> checkedType) {
    if (ct != null) {
      final InterfaceDefinition i = ((InterfaceDefinitionContainer) ct)
          .getInterfaceDefinition();
      final RecordDefinition r = ((RecordDefinitionContainer) ct)
          .getRecordDefinition();
      if (i == null && r == null) fail();
      if (i != null && r != null) fail();

      if (i != null) {
        if (!i.getName().endsWith(ct.getName())) fail();
        checkInterface(i, checkedType);
      } else {
        if (!r.getName().endsWith(ct.getName())) fail();
        checkRecord(r, checkedType);
      }
    }
  }
}
