/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.idl.parser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.InputStream;

import org.junit.Test;
import org.objectweb.fractal.adl.xml.XMLNodeFactory;
import org.objectweb.fractal.adl.xml.XMLNodeFactoryImpl;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Import;
import org.objectweb.fractal.cecilia.adl.idl.ast.ImportContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.Package;
import org.objectweb.fractal.cecilia.adl.idl.ast.PackageContainer;

/**
 * @author Alessio Pace
 */
public class IDLParserTest {

	private IDLParser idlParser;

	/** Path to the DTD defining the IDL AST grammar. */
	public static final String IDL_DTD = "classpath://org/objectweb/fractal/cecilia/adl/idl/ast/cecilia-idl.dtd";

	public final static String INTERFACE_WITH_PACKAGE_PATH = "/org/objectweb/fractal/cecilia/adl/idl/parser/Foo.idl";
	public final static String RECORD_WITH_PACKAGE_PATH = "/org/objectweb/fractal/cecilia/adl/idl/parser/Bar.idl";
	public final static String INTERFACE_WITHOUT_PACKAGE_PATH = "/I1.idl";
	public final static String RECORD_WITHOUT_PACKAGE_PATH = "/R1.idl";

	private XMLNodeFactory nodeFactoryItf = new XMLNodeFactoryImpl();

	protected IDLDefinition parseIDLDefinition(String path) throws Exception {

		InputStream is = IDLParserTest.class.getResourceAsStream(path);

		if (is == null) {
			throw new RuntimeException("Unable to find the IDL file " + path);
		}

		// create the IDLParser
		this.idlParser = new IDLParser(is);
		this.idlParser.init(this.nodeFactoryItf, IDL_DTD, new File(path)
				.getPath());

		// get the IDLDefinition node result which will be used in all the tests
		IDLDefinition definition = this.idlParser.getIDLNode();

		return definition;
	}

	@Test
	public void testImports() throws Exception {

		IDLDefinition definition = parseIDLDefinition(INTERFACE_WITH_PACKAGE_PATH);

		assertTrue("Assert it is an ImportContainer",
				definition instanceof ImportContainer);

		Import[] imports = ((ImportContainer) definition).getImports();
		assertNotNull("Assert Import[] not null", imports);
		assertEquals("Assert number of imports", 0, imports.length);
	}

	@Test
	public void testInterfaceWithNotNullPackage() throws Exception {

		IDLDefinition definition = parseIDLDefinition(INTERFACE_WITH_PACKAGE_PATH);

		assertTrue("Assert it is a PackageContainer",
				definition instanceof PackageContainer);

		Package packageNode = ((PackageContainer) definition).getPackage();
		assertNotNull("Assert Package node not null", packageNode);
		// because the .idl file is in the same package of this test class
		assertEquals("Assert package value", this.getClass().getPackage()
				.getName(), packageNode.getName());
	}

	@Test
	public void testInterfaceWithNullPackage() throws Exception {
		IDLDefinition definition = parseIDLDefinition(INTERFACE_WITHOUT_PACKAGE_PATH);

		assertTrue("Assert it is a PackageContainer",
				definition instanceof PackageContainer);

		Package packageNode = ((PackageContainer) definition).getPackage();
		// because the .idl file has no package declaration
		assertNull("Assert Package node not null", packageNode);

	}

	@Test
	public void testRecordWithNotNullPackage() throws Exception {

		IDLDefinition definition = parseIDLDefinition(RECORD_WITH_PACKAGE_PATH);

		assertTrue("Assert it is a PackageContainer",
				definition instanceof PackageContainer);

		Package packageNode = ((PackageContainer) definition).getPackage();
		assertNotNull("Assert Package node not null", packageNode);
		// because the .idl file is in the same package of this test class
		assertEquals("Assert package value", this.getClass().getPackage()
				.getName(), packageNode.getName());
	}

	@Test
	public void testRecordWithNullPackage() throws Exception {
		IDLDefinition definition = parseIDLDefinition(RECORD_WITHOUT_PACKAGE_PATH);

		assertTrue("Assert it is a PackageContainer",
				definition instanceof PackageContainer);

		Package packageNode = ((PackageContainer) definition).getPackage();
		// because the .idl file has no package declaration
		assertNull("Assert Package node not null", packageNode);

	}

}
