/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Compiler;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.FractalADLTaskMap;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.util.Printer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

import org.objectweb.fractal.task.core.TaskMap;
import org.objectweb.fractal.task.core.Scheduler;
import org.objectweb.fractal.task.core.TaskExecutionException;

/**
 * Implementation of the {@link Factory} interface.
 * 
 * @see org.objectweb.fractal.adl.BasicFactory
 */
public class BasicFactory implements Factory, BindingController {

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /** The name of the {@link Loader} client interface of this component. */
  public static final String LOADER_ITF_NAME    = "loader";

  /** The name of the {@link Compiler} client interface of this component. */
  public static final String COMPILER_ITF_NAME  = "compiler";

  /** The name of the {@link Scheduler} client interface of this component. */
  public static final String SCHEDULER_ITF_NAME = "scheduler";

  /** The {@link Loader} used by this factory. */
  private Loader             loaderItf;

  /** The {@link Compiler} used by this factory. */
  private Compiler           compilerItf;

  /** The {@link Scheduler} used by this factory. */
  private Scheduler          schedulerItf;

  // --------------------------------------------------------------------------
  // Implementation of the Factory interface
  // --------------------------------------------------------------------------

  // Suppress unchecked warning to avoid to change Factory interface
  @SuppressWarnings("unchecked")
  public Object newComponentType(final String name, final Map context)
      throws ADLException {
    throw new UnsupportedOperationException();
  }

  // Suppress unchecked warning to avoid to change Factory interface
  @SuppressWarnings("unchecked")
  public Object newComponent(final String name, final Map context)
      throws ADLException {
    if (Printer.isDebugEnable())
      Printer.debug("[basic-factory] Load ADL \'" + name + "\'");

    final Definition d = loaderItf.load(name, context);
    Printer.debug("[basic-factory] ADL loaded successfully, build task graph");

    final TaskMap m = new FractalADLTaskMap();
    compilerItf.compile(d, m, context);
    Printer
        .debug("[basic-factory] Task graph build successfully, execute tasks");
    
    try{
      schedulerItf.schedule(m.getTasks(), context);
    }catch(TaskExecutionException tee){
      //tee.printStackTrace();
      throw new ADLException("Unable to execute one of the tasks", tee);
    }
    Printer.debug("[basic-factory] Tasks executed successfully.");
    final Object result = m.getTask("output", d).getResult();
    return result;
  }

  public void bindFc(final String itfName, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    // XXX check for value nullness?

    try {
      if (itfName.equals(LOADER_ITF_NAME)) {
        this.loaderItf = (Loader) value;
      } else if (itfName.equals(COMPILER_ITF_NAME)) {
        this.compilerItf = (Compiler) value;
      } else if (itfName.equals(SCHEDULER_ITF_NAME)) {
        this.schedulerItf = (Scheduler) value;
      } else {
        throw new NoSuchInterfaceException(
            "Unable to find the interface named '" + itfName + "'");
      }
    } catch (final ClassCastException cce) {
      cce.printStackTrace();
      throw new IllegalBindingException("Unmatching interface types");
    }
  }

  public String[] listFc() {
    return new String[]{LOADER_ITF_NAME, COMPILER_ITF_NAME, SCHEDULER_ITF_NAME};
  }

  public Object lookupFc(final String itfName) throws NoSuchInterfaceException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(LOADER_ITF_NAME)) {
      return this.loaderItf;
    } else if (itfName.equals(COMPILER_ITF_NAME)) {
      return this.compilerItf;
    } else if (itfName.equals(SCHEDULER_ITF_NAME)) {
      return this.schedulerItf;
    } else {
      throw new NoSuchInterfaceException("Unable to find the interface named '"
          + itfName + "'");
    }
  }

  public void unbindFc(final String itfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(LOADER_ITF_NAME)) {
      this.loaderItf = null;
    } else if (itfName.equals(COMPILER_ITF_NAME)) {
      this.compilerItf = null;
    } else if (itfName.equals(SCHEDULER_ITF_NAME)) {
      this.schedulerItf = null;
    } else {
      throw new NoSuchInterfaceException("Unable to find the interface named '"
          + itfName + "'");
    }

  }
}
