/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 * Contributor: Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.attributes;

import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.attributes.Attribute;
import org.objectweb.fractal.adl.attributes.Attributes;
import org.objectweb.fractal.adl.attributes.AttributesContainer;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.components.AbstractPrimitiveCompiler;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinitionContainer;
import org.objectweb.fractal.task.codegen.api.RequireSourceCodeProviderTask;
import org.objectweb.fractal.task.codegen.lib.AbstractRecursiveSourceCodeProviderTask;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Adds a task to set the attributes of a component.
 */
public class AttributeCompiler extends AbstractPrimitiveCompiler
    implements
      BindingController {

  // --------------------------------------------------------------------------
  // Client interface.
  // --------------------------------------------------------------------------

  /**
   * The name of the {@link AttributeBuilder} client interface used by this
   * compiler.
   */
  public static final String BUILDER_ITF = "builder";

  /** The {@link AttributeBuilder} used by this compiler. */
  protected AttributeBuilder builderItf;

  // --------------------------------------------------------------------------
  // Implementation of the Compiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {
    Attributes attributes = null;
    if (container instanceof AttributesContainer) {
      attributes = ((AttributesContainer) container).getAttributes();
    }
    if (attributes == null) return;

    createTask(container, tasks, attributes);
  }

  protected void createTask(final ComponentContainer container,
      final TaskMap tasks, final Attributes attributes) {
    final AttributesTask task = new AttributesTask(builderItf,
        ((RecordDefinitionContainer) attributes).getRecordDefinition(),
        attributes.getAttributes());
    final TaskMap.TaskHole taskHole = addTask("attributes", task, container,
        tasks);
    getTaskHole("component", container, tasks).addDependency(taskHole,
        RequireSourceCodeProviderTask.SOURCE_CODE_PROVIDER_TASK_ROLE,
        "attributes");
  }

  // --------------------------------------------------------------------------
  // Task class
  // --------------------------------------------------------------------------

  protected static class AttributesTask
      extends
        AbstractRecursiveSourceCodeProviderTask {
    private final Object           attributeAST;
    private final AttributeBuilder builder;
    private final Attribute[]      fields;

    /**
     * @param builder
     * @param attributeAST
     * @param fields
     */
    public AttributesTask(final AttributeBuilder builder,
        final Object attributeAST, final Attribute[] fields) {
      this.builder = builder;
      this.attributeAST = attributeAST;
      this.fields = fields;
    }

    public void execute(final Map<Object, Object> context) throws Exception {
      final Object factory = (getFactoryProviderTask() == null)
          ? null
          : getFactoryProviderTask().getFactory();
      final Object instance = (getInstanceProviderTask() == null)
          ? null
          : getInstanceProviderTask().getInstance();

      setResult(builder.setAttributes(factory, instance, attributeAST, fields,
          getCodePieces(), context));
    }
  }

  public void bindFc(final String s, final Object o)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = (AttributeBuilder) o;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }

  }

  public String[] listFc() {
    return new String[]{BUILDER_ITF};
  }

  public Object lookupFc(final String s) throws NoSuchInterfaceException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      return this.builderItf;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

  public void unbindFc(final String s) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (s == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (s.equals(BUILDER_ITF)) {
      this.builderItf = null;
    } else {
      throw new NoSuchInterfaceException("There is no interface named '" + s
          + "'");
    }
  }

}
