/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 */

package org.objectweb.fractal.cecilia.adl.idl;

import static org.objectweb.fractal.adl.NodeUtil.castNodeError;
import static org.objectweb.fractal.cecilia.adl.idl.util.Util.getImportedAST;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.ContextLocal;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.attributes.Attributes;
import org.objectweb.fractal.adl.attributes.AttributesContainer;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.components.PrimitiveCompiler;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.Import;
import org.objectweb.fractal.cecilia.adl.idl.ast.ImportContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinitionContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.RecordDefinitionContainer;
import org.objectweb.fractal.task.core.TaskMap;

/**
 * Travels the component's interfaces and attribute definitions and invokes
 * recursively its client IDL compiler interfaces for all these definitions and
 * all the definitions which are referenced.
 */
public class RecursiveIDLDispatcher
    implements
      PrimitiveCompiler,
      BindingController {

  /** A local set to store already compiled IDL definitions. */
  private final ContextLocal<Set<String>> compiledIDLDefinitions = new ContextLocal<Set<String>>();

  // --------------------------------------------------------------------------
  // Client interfaces
  // --------------------------------------------------------------------------

  /** The name of the builder collection interface. */
  public static final String              COMPILER_BINDING       = "client-compiler";
  /** The builders client interfaces. */
  public Map<String, IDLCompiler>         compilersItf           = new HashMap<String, IDLCompiler>();

  // --------------------------------------------------------------------------
  // Implementation of the PrimitiveCompiler interface
  // --------------------------------------------------------------------------

  public void compile(final List<ComponentContainer> path,
      final ComponentContainer container, final TaskMap tasks,
      final Map<Object, Object> context) throws ADLException {
    // invoke the set of IDL compilers that are connected
    if (container instanceof InterfaceContainer) {
      for (final Interface itf : ((InterfaceContainer) container)
          .getInterfaces()) {
        invokeCompilers(path, container, castNodeError(itf,
            InterfaceDefinitionContainer.class).getInterfaceDefinition(),
            tasks, context);
      }
    }
    if (container instanceof AttributesContainer) {
      final Attributes attributes = ((AttributesContainer) container)
          .getAttributes();
      if (attributes != null)
        invokeCompilers(path, container, castNodeError(attributes,
            RecordDefinitionContainer.class).getRecordDefinition(), tasks,
            context);
    }
  }

  // --------------------------------------------------------------------------
  // Utility method
  // --------------------------------------------------------------------------

  private void invokeCompilers(final List<ComponentContainer> path,
      final ComponentContainer container, final IDLDefinition idlDefinition,
      final TaskMap tasks, final Map<Object, Object> context)
      throws ADLException {
    final Set<String> cache = getCache(context);

    // add the IDLDefinition's name in cache. If it is not added (i.e. it is
    // already present in the set), then return.
    if (!cache.add(idlDefinition.getName())) return;

    // Invoke all the client idl-compiler interfaces for this IDL definition
    for (final IDLCompiler idlCompiler : compilersItf.values()) {
      idlCompiler.compile(path, idlDefinition, container, tasks, context);
    }
    if (idlDefinition instanceof ImportContainer) {
      for (final Import imp : ((ImportContainer) idlDefinition).getImports()) {
        final Node importedAST = getImportedAST(imp);
        if (importedAST instanceof IDLDefinition)
          invokeCompilers(path, container, (IDLDefinition) importedAST, tasks,
              context);
      }
    }
  }

  private Set<String> getCache(final Map<Object, Object> context) {
    Set<String> cache = compiledIDLDefinitions.get(context);
    if (cache == null) {
      cache = new HashSet<String>();
      compiledIDLDefinitions.set(context, cache);
    }
    return cache;
  }

  // --------------------------------------------------------------------------
  // Implementation of the BindignController interface
  // --------------------------------------------------------------------------

  public String[] listFc() {
    return compilersItf.keySet().toArray(new String[compilersItf.size()]);
  }

  public Object lookupFc(final String itf) throws NoSuchInterfaceException {
    if (itf.startsWith(COMPILER_BINDING)) {
      return compilersItf.get(itf);
    }
    return null;
  }

  public void bindFc(final String itf, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {
    if (itf.startsWith(COMPILER_BINDING)) {
      compilersItf.put(itf, (IDLCompiler) value);
    }
  }

  public void unbindFc(final String itf) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {
    if (itf.startsWith(COMPILER_BINDING)) {
      compilersItf.remove(itf);
    }
  }

}
